/*
** $Id: whokill.c,v 1.3 2000/02/27 18:10:27 jtison Stable $
** $Log: whokill.c,v $
** Revision 1.3  2000/02/27 18:10:27  jtison
** Updated for xchat 1.4.0
** - Moved all globally scoped data objects to static storage
**   to prevent nasty SIGSEGVs.
** - Removed Adam Longley's mmap-based memory allocator. Looks
**   like xchat 1.4.0 fixed the heap walker.
** - Vastly extended comments.
**
** Revision 1.2  1999/11/27 20:11:46  jtison
** Properly stringify nick list.
** Stop collecting at MAX_NICKS.
** Implement -v (version) command switch.
**
** Revision 1.1  1999/11/26 19:09:18  jtison
** Initial revision
**
*/

#ifndef USE_PLUGINS
#define USE_PLUGINS
#endif

static const char *RCS_VERSION_ID_WHOKILL = "$Id: whokill.c,v 1.3 2000/02/27 18:10:27 jtison Stable $ - (c) 1999 J. S. Tison";

/* whokill - Standard IRC operator tool. Written as a familiarity adventure
 *           with xchat plugin C code. Some of the remarks you see in here
 *           come from observations. I offer this code as a learning experience
 *           for other xchat C programmers. Unfortunately, there isn't much 
 *           documentation about writing plugins in C. 
 *
 *	WARNING: This plugin was originally written for xchat 1.3.7, 
 *	which apparently had some kind of heap corruption bug. It has
 *	since been updated for xchat 1.4.0, and is ONLY guaranteed to
 *	work with 1.4.0 (and possibly higher, if the corruption bug 
 *	isn't re-introduced). For a brilliant (but piggish) workaround,
 *	refer to Adam Longley's 'Auto' plugin, and slap in his mmap_malloc()
 *	and mmap_free() routines. Being able to use the *real* heap, 
 *	however, is MUCH more efficient, and I've since removed Adam's 
 *	routines. 
 *
 *  Copyright (C) 1999, 2000 Jim Tison
 *  November, 1999 ... last update February, 2000. 
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */ 

#include <xchat.h>
#include <plugin.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <stdio.h>
#include <ctype.h>

/*
**  xchat externals (YOY isn't there a header file for these?)
*/

extern struct module  *module_find (char *name);
extern int            tcp_send (struct server *serv, char *buf);
extern void           PrintText (struct session *sess, char *text);
extern void	      hook_signal(struct xp_signal *);
extern void	      module_add_cmds(struct module_cmd_set *); 

/*
**  Internal function prototypes
*/

int         module_init (int ver, struct module *mod, struct session *sess);
int         wk_filter(struct session *sess, char *server, char *text, char *null1, char *null2, char null3);
static void wk_send_kills(struct session *sess);
static void wk_collect(struct session *sess, char *whoinfo);
int         wk_handle_command(struct session *sess, char *tbuf, char *word[], char *word_eol[]);
static void sPrintText(struct session *sess, char *fmt, ...);
static void sendRaw(struct session *sess, char *fmt, ...);

/*
**  Custom externals    
*/

char    *name = "WhoKill";
char    *desc = "IRC oper WHO utilities";

static struct xp_signal	inbound_sig;
int                   	(*inbound_next)(void *, void *, void *, void *, void *, char);

struct commands		wk_commands[] = { 
                        { "WHOKILL",    wk_handle_command,    0, 0, "/WHOKILL <user@host> <comment>" },   
                        { NULL,         0,                    0, 0, 0 }
};

static struct module_cmd_set 	cmd_set;

/*
**  Global custom locals 
*/
#define	MAX_NICKS	400
static char *   nick[MAX_NICKS];
static int      nicks;        
static char *   comment;
static int      min_for_kill;
static int      trapping;
static int      listonly;

/*
 *	module_init() is a required function. It is called as a direct
 *	result of the xchat LOADDLL command. In this routine, we make
 *	sure that: 
 *		- We are compatible (coded to) the current plugin
 *		  interface standard
 *		- This module isn't loaded more than once
 *		- We hook the XP_INBOUND signal to force xchat to 
 *		  call our main routine whenever an inbound message
 *		  (IRC traffic) is received
 *		- We register our command set and associate it with
 *		  this module. 
 *		- We perform whatever other initialization tasks are
 *		  necessary (set up initial data objects, etc).
 */


int     module_init (int ver, struct module *mod, struct session *sess) { 
  if (ver != MODULE_IFACE_VER)		/* This MUST be the first test!!	*/
    return 1;
  if (module_find(name) != NULL) { 	/* Don't allow us to load twice.	*/
    PrintText(sess, "Module WhoKill already loaded\n");
    return 1;
  }
  PrintText(sess, "Module WhoKill is loaded\n");
  mod->name = name;  			/* Identify ourselves               	*/
  mod->desc = desc;                     /* Give purpose of module           	*/

  /*
  **    We are hooking only the server inbound message. It goes to 
  **    a filter (wk_filter) that acts only on 352 (who) or 315 (end of who)
  **    numeric messages. Everything else gets passed along to the next
  **	hook.
  */
  inbound_sig.signal = XP_INBOUND;		/* On every incoming message,	*/
  inbound_sig.callback = XP_CALLBACK(wk_filter); /* call wk_filter.		*/   	
  inbound_sig.naddr = &inbound_next;		/* Placeholder for next hook.	*/
  inbound_sig.mod = mod; 			/* Associate signal w/this mod.	*/

  /*    Hook the XP_INBOUND signal      */

  hook_signal(&inbound_sig);			/* Register the hook.		*/
  cmd_set.mod = mod;				
  cmd_set.cmds = wk_commands;
  module_add_cmds(&cmd_set);			/* Register our commands.	*/

  /*    Initialize module-wide variables to the defaults   */

  nicks = 0;                      
  min_for_kill = 4;
  return 0;
}

/*
**    This function is the callback gate for whokill. The <text> argument
**    is a pointer to the raw server text buffer. If this isn't a numeric
**    server message 315 or 352, just pass it on to the next event in the
**    chain. Otherwise, see if we're active.
*/
int     wk_filter(struct session *sess, char *server, char *text, char *null1, char *null2, char null3) {
  char        *p[4] = { NULL, NULL, NULL, NULL };
  int         i, msgnbr;

  /*  Find the first 3 spaces                                                 */

  for (p[0]=text, i = 1; i < 4; i++) { 
    p[i] = strchr(p[i-1], ' ');
    if (!p[i])                              /* Don't try to execute strchr()  */
      break;                                /*  with a NULL string base!      */
  }
  /*
  **    We have a 2nd token in the text. Try to make it a number. 
  */
  if(p[2]) {
    msgnbr = atoi(p[2]+1);
    if(msgnbr == 315 || msgnbr == 352) {    /* Only work 315/352                */
      if (trapping) {                       /* Don't care unless we're trapping */
        if(msgnbr == 352)
          wk_collect(sess, (p[3]+1));
        else
          wk_send_kills(sess);
      }                                     /* End wk_trapping                  */
    }                                       /* End 315/352                      */
  }                                         /* End 2nd token.                   */
  XP_CALLNEXT(inbound_next, sess, server, text, null1, null2, null3); /* Pass   */
}   

        
static void     wk_send_kills(struct session *sess) { 
  int             	i;
  char			*buf;
  char			work[40];

  trapping = 0;
  /*	
  **	If listonly, DO NOT KILL. Just print back the matches.
  */
  if (listonly) { 
    buf = malloc(17*nicks + 13 + strlen(comment));
    strcpy(buf, comment);
    sprintf(work, " [count=%d] ", nicks);
    strcat(buf, work);
    for(i=0; i < nicks; i++) { 
      if(!i) 
	strcat(buf, nick[0]);
      else { 
	strcat(buf, ",");
	strcat(buf, nick[i]);
      }
    }
    sPrintText(sess, "%s\n", buf);	/* Produce the display line.	*/
    free(buf);				/* Return our work buffer.	*/
    listonly = 0;			/* Reset whokill state.		*/
  }
  /*
  **	Not a listonly. Do the KILLs. 
  */
  else {
    if(nicks < min_for_kill) {
      sPrintText(sess, "%s: minimum of %d matches from WHO not met(%d). Aborting request.\n",
		 name, min_for_kill, nicks); 
    }
    else for(i=0; i < nicks; i++) 
      sendRaw(sess, "KILL %s :%s <%d/%d>\n", nick[i], comment, i+1, nicks); 
  }
  /*
  **	Clean up the heap memory items and reset state for next.
  */
  for(i=0; i < nicks; i++)
    free(nick[i]);
  free(comment);
  nicks = 0;
  return;
}

/*
**	module_cleanup(): 
**		This is a required routine, and is activated directly
**		by the UNLOADDLL command. If you have any final 
**		clean up tasks to accomplish, do them here. This 
**		example has none, so it just prints an info 
**		message.
*/
  
void          module_cleanup(struct module *mod, struct session *sess) { 
  PrintText(sess, "Module WhoKill unloading\n");
}


static void   wk_collect(struct session *sess, char *whoinfo) { 
  int           i; 
  char          *m, *s, *e;
  char 					*temp;
  
  /*  
  **  The token we have passed as <*whoinfo> looks like this, and it may  
  **  NOT be tokenized with strtok():                   
  **    me channel userid host server nickname umodes :hopcount realname
  **      ^       ^      ^    ^      ^        ^
  **   e: 0       1      2    3      4        5  
  */
  if(nicks == MAX_NICKS - 1) { 
    sPrintText(sess, "%s: This nick (# %d) is the last I can hold. Reset MAX_NICKS.\n",
	       name, nicks+1);
  }
  if (nicks == MAX_NICKS)
    return;

  m = strchr(whoinfo, ':');           /* Find hopcount token delimiter        */
  if(!m) { 
    PrintText(sess, "Unexpected return in WHO buffer!! Check results!\n");
    return;
  }
  for(i=0, e = whoinfo; i < 7; i++) { 
    s = e + 1; 
    e = strchr(s, ' ');
  }
  i = e - s;		
  temp = malloc(i+1);                 /* Leave room for terminating zero      */
  strncpy(temp, s, i);	   	      /* Copy the string data                 */ 
  *(temp+i) = 0;                      /* Terminate the string                 */
  nick[nicks++] = temp;               /* Anchor the collected nickname.       */
  return;                             /* We're done.                          */ 
}

#define HELP_LINES  13

const char *  wk_help_text[HELP_LINES] = { 
  "--WhoKill HELP--\n",
  " usage: WHOKILL [-t number][-h][-r][-m][-v] userhost_mask comments\n",
  "   -t number:  Set the minimum kill threshold. Default is 4. WhoKill will\n",
  "               not issue any kills if the number of matches from WHO is\n",
  "               less than this number. -t1 means kill any match, but\n",  
  "               -t0 is an error.\n",
  "   -h:         Display this help.\n",
  "   -r:         Report total matches to <userhost_mask>. DO NOT KILL\n",
  "   -m:         Display matching mask in comment, following comment text.\n",
  "   -v:         Display version information.\n",
  " userhost_mask The userhost mask you want to kill. BE CAREFUL. It IS\n", 
  "               possible to kill * (all matches, including yourself)!!\n",
  " comments      Comment to go on the KILL command. THIS IS REQUIRED.\n"
};


/*
**	This is the main command handler. 
**		word[] is an array of char pointers containing tokenized
**			(space delimited) words of the command string. word[0] is the
**			server, if any, to which you are connected (I think??) IT IS 
**			UNDEPENDABLE, and may point to undisplayable garbage. word[1]
**		        is the name of the command as entered. word[2] is the first 
**			token of the command arguments... word[31] is the last available,
**			(29th command token) regardless of how many tokens there are.
**	 	word_eol[] is like word[], except that each token continues to 
**			the end of the command line.
**		The first array member containing nothing points to a null
**			string (constant char 0 somewhere in the text segment). You
**			must test each with strlen() to know if its present. 
**		tbuf's contents are UNKNOWN and UNDEPENDABLE. 
*/  
int           wk_handle_command(struct session *sess, char *tbuf, char *word[], char *word_eol[]) {
  int           i, j;
  int		save_mask = 0;
  char 		*t;				

#ifndef DEBUG
  if (sess->server->connected == 0) { 
    sPrintText(sess, "%s: Sorry -- you are not connected. Command ignored.\n", name);
    return 0;
  }
#endif

  if(trapping) { 
    sPrintText(sess, "%s: Sorry - another %s is already in progress\n", name, name);
    return 0;
  }

  min_for_kill = 4;
  nicks = 0;

  /*
  **	Parse out any command switches
  */

  for(i=2; i < 6; i++) { 
    if(*word[i] != '-')
      break;
    else { 
      t = word[i]+1;
      switch(tolower(*t)) { 
      case 't':
	j = atoi(t+1);
	if(!j) { 
	  sPrintText(sess, "%s: Illegal threshold value. Ignoring this command.\n", name);
	  return 0;
	}
	min_for_kill = j;				
	break;
      case 'r':
	listonly = 1;			/* FALLTHRU: -r implies -m.		*/
      case 'm':		
	save_mask = 1;
	break;
      case 'v':
	sPrintText(sess, "%s\n", RCS_VERSION_ID_WHOKILL);
	return 0;
      case 'h':
	for(i=0; i < HELP_LINES; i++)
	  PrintText(sess, (char *)wk_help_text[i]);
	return 0;
      default:
	sPrintText(sess, "%s: Unrecognized command switch. See /WHOKILL -h for usage\n",
		   name);
	return 0;
      }
    }
  }
  if (!strlen(word[i])) { 
    sPrintText(sess, "%s: Missing search (user@)host mask. Command ignored.\n", name);
    return 0;
  }
  /*
  **	Resolve the comment. If doing a normal kill, the field is required 
  **	from the command line. If not, it equals "Nicks". 
  */
  if(!listonly && !strlen(word[i+1])) {
    sPrintText(sess, "%s: Missing KILL comment. Command ignored.\n", name);
    return 0;
  }
  else if(listonly)
    t = strdup("Nicks");
  else 					/* !listonly && comment exists		*/ 
    t = strdup(word_eol[i+1]);

  if(!save_mask)
    comment = t;			/* not matching the mask in comment	*/
  else { 								
    comment = malloc(strlen(t) + strlen(word[i]) + 25);
    sprintf(comment, "%s (matching %s)", t, word[i]);
    free(t);
  }
  
  trapping = 1;
  sendRaw(sess, "WHO %s\n", word[i]);
  return 0;
} 


/*
**	These two utility routines, sPrintText() and sendRaw(),
**	are programmer-assist routines. 
**
**	sPrintText() works like v<x>printf(), allowing a va_list
**	set of arguments to fill out a message.
**
**	sendRaw() forces the output of a raw IRC command to the 
**	server uplink. 
*/

#define	sPRINTTEXT_BUFSIZE	512
#define sPRINTTEXT_LIMIT	511   
static void	sPrintText(struct session *sess, char *fmt, ...) { 
  char		  buffer[sPRINTTEXT_BUFSIZE];
  va_list	  vargs;

  va_start(vargs, fmt);
  vsnprintf(buffer, sPRINTTEXT_LIMIT, fmt, vargs);
  PrintText(sess, buffer);
  return;
}
	

static void	sendRaw(struct session *sess, char *fmt, ...) { 
  va_list	  vargs;	
  char		  buffer[sPRINTTEXT_BUFSIZE];
  
  va_start(vargs, fmt);
  vsnprintf(buffer, sPRINTTEXT_LIMIT, fmt, vargs);
  tcp_send(sess->server, buffer);
  return;
}
